# Generated by Django 2.2.5 on 2019-09-16 12:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('knox', '0007_auto_20190111_0542'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authtoken',
            name='salt',
        ),
    ]
