# Generated by Django 4.1rc1 on 2022-07-20 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("knox", "0008_remove_authtoken_salt"),
    ]

    operations = [
        migrations.AlterField(
            model_name="authtoken",
            name="token_key",
            field=models.CharField(db_index=True, max_length=25),
        ),
    ]
